/*
 * Copyright 2016 - 2022 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#pragma once

#include <filesystem>
#include <fstream>

#include <ox/model/typenamecatcher.hpp>
#include <ox/oc/oc.hpp>
#include <ox/std/buffer.hpp>
#include <ox/std/defines.hpp>
#include <ox/std/fmt.hpp>
#include <ox/std/trace.hpp>
#include <ox/std/string.hpp>

#include <nostalgia/core/context.hpp>

namespace nostalgia::studio {

constexpr auto ConfigDir = [] {
	switch (ox::defines::OS) {
		case ox::OS::Darwin:
			return "{}/Library/Preferences/{}";
		case ox::OS::DragonFlyBSD:
		case ox::OS::FreeBSD:
		case ox::OS::Linux:
		case ox::OS::NetBSD:
		case ox::OS::OpenBSD:
			return "{}/.config/{}";
		case ox::OS::BareMetal:
		case ox::OS::Windows:
			return "";
	}
}();

template<typename T>
ox::Result<T> readConfig(core::Context *ctx, const ox::String &name) noexcept {
	oxAssert(name != "", "Config type has no TypeName");
	const auto homeDir = std::getenv("HOME");
	const auto configPath = ox::sfmt(ConfigDir, homeDir, ctx->appName).toStdString();
	const auto path = ox::sfmt("{}/{}.json", configPath, name).toStdString();
	std::ifstream file(path, std::ios::binary | std::ios::ate);
	if (!file.good()) {
		oxErrf("Could not find config file: {}\n", path);
		return OxError(1, "Could not find config file");
	}
	try {
		const auto size = file.tellg();
		file.seekg(0, std::ios::beg);
		ox::Buffer buff(static_cast<std::size_t>(size));
		file.read(buff.data(), size);
		return ox::readOC<T>(buff);
	} catch (const std::ios_base::failure &e) {
		oxErrf("Could not read config file: {}\n", e.what());
		return OxError(2, "Could not read config file");
	}
}

template<typename T>
ox::Result<T> readConfig(core::Context *ctx) noexcept {
	constexpr auto TypeName = ox::requireModelTypeName<T>();
	return readConfig<T>(ctx, TypeName);
}

template<typename T>
ox::Error writeConfig(core::Context *ctx, const ox::String &name, T *data) noexcept {
	oxAssert(name != "", "Config type has no TypeName");
	const auto homeDir = std::getenv("HOME");
	const auto configPath = ox::sfmt(ConfigDir, homeDir, ctx->appName).toStdString();
	const auto path = ox::sfmt("{}/{}.json", configPath, name).toStdString();
	std::error_code ec;
	std::filesystem::create_directory(configPath, ec);
	if (ec) {
		oxErrf("Could not create config directory: {}\n", ec.message());
	}
	std::ofstream file(path, std::ios::binary | std::ios::ate);
	if (!file.good()) {
		oxErrf("Could not find config file: {}\n", path);
		return OxError(1, "Could not find config file");
	}
	oxRequireM(buff, ox::writeOC(data));
	buff.back().value = '\n';
	try {
		file.write(buff.data(), static_cast<ox::Signed<decltype(buff.size())>>(buff.size()));
		return OxError(0);
	} catch (const std::ios_base::failure &e) {
		oxErrf("Could not read config file: {}\n", e.what());
		return OxError(2, "Could not read config file");
	}
}

template<typename T>
ox::Error writeConfig(core::Context *ctx, T *data) noexcept {
	constexpr auto TypeName = ox::requireModelTypeName<T>();
	return writeConfig(ctx, TypeName, data);
}

template<typename T, typename Func>
void openConfig(core::Context *ctx, const ox::String &name, Func f) noexcept {
	oxAssert(name != "", "Config type has no TypeName");
	const auto c = readConfig<T>(ctx, name);
	f(&c.value);
}

template<typename T, typename Func>
void openConfig(core::Context *ctx, Func f) noexcept {
	constexpr auto TypeName = ox::requireModelTypeName<T>();
	openConfig<T>(ctx, TypeName, f);
}

template<typename T, typename Func>
void editConfig(core::Context *ctx, const ox::String &name, Func f) noexcept {
	oxAssert(name != "", "Config type has no TypeName");
	auto c = readConfig<T>(ctx, name);
	f(&c.value);
	oxLogError(writeConfig(ctx, name, &c.value));
}

template<typename T, typename Func>
void editConfig(core::Context *ctx, Func f) noexcept {
	constexpr auto TypeName = ox::requireModelTypeName<T>();
	editConfig<T>(ctx, TypeName, f);
}

}
